@@ -18,6 +18,7 @@ import (
 	"encoding/binary"
 
 	"errors"
+	"hash/crc32"
 	"io/ioutil"
 	"math/rand"
 	"os"
@@ -26,8 +27,9 @@ import (
 	"testing"
 
 	"github.com/go-kit/kit/log"
+	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/tsdb/chunks"
-	"github.com/prometheus/prometheus/tsdb/labels"
+	"github.com/prometheus/prometheus/tsdb/fileutil"
 	"github.com/prometheus/prometheus/tsdb/tsdbutil"
 	"github.com/prometheus/prometheus/util/testutil"
 )
@@ -87,14 +89,16 @@ func TestCreateBlock(t *testing.T) {
 func TestCorruptedChunk(t *testing.T) {
 	for name, test := range map[string]struct {
 		corrFunc func(f *os.File) // Func that applies the corruption.
-		expErr   error
+		openErr  error
+		queryErr error
 	}{
 		"invalid header size": {
 			func(f *os.File) {
 				err := f.Truncate(1)
 				testutil.Ok(t, err)
 			},
-			errors.New("invalid chunk header in segment 0: invalid size"),
+			errors.New("invalid segment header in segment 0: invalid size"),
+			nil,
 		},
 		"invalid magic number": {
 			func(f *os.File) {
@@ -110,6 +114,7 @@ func TestCorruptedChunk(t *testing.T) {
 				testutil.Equals(t, chunks.MagicChunksSize, n)
 			},
 			errors.New("invalid magic number 0"),
+			nil,
 		},
 		"invalid chunk format version": {
 			func(f *os.File) {
@@ -125,6 +130,45 @@ func TestCorruptedChunk(t *testing.T) {
 				testutil.Equals(t, chunks.ChunksFormatVersionSize, n)
 			},
 			errors.New("invalid chunk format version 0"),
+			nil,
+		},
+		"chunk not enough bytes to read the chunk length": {
+			func(f *os.File) {
+				// Truncate one byte after the segment header.
+				err := f.Truncate(chunks.SegmentHeaderSize + 1)
+				testutil.Ok(t, err)
+			},
+			nil,
+			errors.New("segment doesn't include enough bytes to read the chunk size data field - required:13, available:9"),
+		},
+		"chunk not enough bytes to read the data": {
+			func(f *os.File) {
+				fi, err := f.Stat()
+				testutil.Ok(t, err)
+
+				err = f.Truncate(fi.Size() - 1)
+				testutil.Ok(t, err)
+			},
+			nil,
+			errors.New("segment doesn't include enough bytes to read the chunk - required:26, available:25"),
+		},
+		"checksum mismatch": {
+			func(f *os.File) {
+				fi, err := f.Stat()
+				testutil.Ok(t, err)
+
+				// Get the chunk data end offset.
+				chkEndOffset := int(fi.Size()) - crc32.Size
+
+				// Seek to the last byte of chunk data and modify it.
+				_, err = f.Seek(int64(chkEndOffset-1), 0)
+				testutil.Ok(t, err)
+				n, err := f.Write([]byte("x"))
+				testutil.Ok(t, err)
+				testutil.Equals(t, n, 1)
+			},
+			nil,
+			errors.New("checksum mismatch expected:cfc0526c, actual:34815eae"),
 		},
 	} {
 		t.Run(name, func(t *testing.T) {
@@ -134,7 +178,8 @@ func TestCorruptedChunk(t *testing.T) {
 				testutil.Ok(t, os.RemoveAll(tmpdir))
 			}()
 
-			blockDir := createBlock(t, tmpdir, genSeries(1, 1, 0, 1))
+			series := newSeries(map[string]string{"a": "b"}, []tsdbutil.Sample{sample{1, 1}})
+			blockDir := createBlock(t, tmpdir, []Series{series})
 			files, err := sequenceFiles(chunkDir(blockDir))
 			testutil.Ok(t, err)
 			testutil.Assert(t, len(files) > 0, "No chunk created.")
@@ -146,8 +191,21 @@ func TestCorruptedChunk(t *testing.T) {
 			test.corrFunc(f)
 			testutil.Ok(t, f.Close())
 
-			_, err = OpenBlock(nil, blockDir, nil)
-			testutil.Equals(t, test.expErr.Error(), err.Error())
+			// Check open err.
+			b, err := OpenBlock(nil, blockDir, nil)
+			if test.openErr != nil {
+				testutil.Equals(t, test.openErr.Error(), err.Error())
+				return
+			}
+
+			querier, err := NewBlockQuerier(b, 0, 1)
+			testutil.Ok(t, err)
+			set, err := querier.Select(labels.MustNewMatcher(labels.MatchEqual, "a", "b"))
+			testutil.Ok(t, err)
+
+			// Check query err.
+			testutil.Equals(t, false, set.Next())
+			testutil.Equals(t, test.queryErr.Error(), set.Err().Error())
 		})
 	}
 }
@@ -175,16 +233,17 @@ func TestBlockSize(t *testing.T) {
 			testutil.Ok(t, blockInit.Close())
 		}()
 		expSizeInit = blockInit.Size()
-		actSizeInit := testutil.DirSize(t, blockInit.Dir())
+		actSizeInit, err := fileutil.DirSize(blockInit.Dir())
+		testutil.Ok(t, err)
 		testutil.Equals(t, expSizeInit, actSizeInit)
 	}
 
 	// Delete some series and check the sizes again.
 	{
-		testutil.Ok(t, blockInit.Delete(1, 10, labels.NewMustRegexpMatcher("", ".*")))
+		testutil.Ok(t, blockInit.Delete(1, 10, labels.MustNewMatcher(labels.MatchRegexp, "", ".*")))
 		expAfterDelete := blockInit.Size()
 		testutil.Assert(t, expAfterDelete > expSizeInit, "after a delete the block size should be bigger as the tombstone file should grow %v > %v", expAfterDelete, expSizeInit)
-		actAfterDelete := testutil.DirSize(t, blockDirInit)
+		actAfterDelete, err := fileutil.DirSize(blockDirInit)
 		testutil.Ok(t, err)
 		testutil.Equals(t, expAfterDelete, actAfterDelete, "after a delete reported block size doesn't match actual disk size")
 
@@ -198,15 +257,50 @@ func TestBlockSize(t *testing.T) {
 			testutil.Ok(t, blockAfterCompact.Close())
 		}()
 		expAfterCompact := blockAfterCompact.Size()
-		actAfterCompact := testutil.DirSize(t, blockAfterCompact.Dir())
+		actAfterCompact, err := fileutil.DirSize(blockAfterCompact.Dir())
+		testutil.Ok(t, err)
 		testutil.Assert(t, actAfterDelete > actAfterCompact, "after a delete and compaction the block size should be smaller %v,%v", actAfterDelete, actAfterCompact)
 		testutil.Equals(t, expAfterCompact, actAfterCompact, "after a delete and compaction reported block size doesn't match actual disk size")
 	}
 }
 
+func TestReadIndexFormatV1(t *testing.T) {
+	/* The block here was produced at commit
+	   07ef80820ef1250db82f9544f3fcf7f0f63ccee0 with:
+	db, _ := Open("v1db", nil, nil, nil)
+	app := db.Appender()
+	app.Add(labels.FromStrings("foo", "bar"), 1, 2)
+	app.Add(labels.FromStrings("foo", "baz"), 3, 4)
+	app.Add(labels.FromStrings("foo", "meh"), 1000*3600*4, 4) // Not in the block.
+	app.Commit()
+	db.compact()
+	db.Close()
+	*/
+
+	blockDir := filepath.Join("testdata", "index_format_v1")
+	block, err := OpenBlock(nil, blockDir, nil)
+	testutil.Ok(t, err)
+
+	q, err := NewBlockQuerier(block, 0, 1000)
+	testutil.Ok(t, err)
+	testutil.Equals(t, query(t, q, labels.MustNewMatcher(labels.MatchEqual, "foo", "bar")),
+		map[string][]tsdbutil.Sample{`{foo="bar"}`: []tsdbutil.Sample{sample{t: 1, v: 2}}})
+
+	q, err = NewBlockQuerier(block, 0, 1000)
+	testutil.Ok(t, err)
+	testutil.Equals(t, query(t, q, labels.MustNewMatcher(labels.MatchNotRegexp, "foo", "^.$")),
+		map[string][]tsdbutil.Sample{
+			`{foo="bar"}`: []tsdbutil.Sample{sample{t: 1, v: 2}},
+			`{foo="baz"}`: []tsdbutil.Sample{sample{t: 3, v: 4}},
+		})
+}
+
 // createBlock creates a block with given set of series and returns its dir.
 func createBlock(tb testing.TB, dir string, series []Series) string {
-	head := createHead(tb, series)
+	return createBlockFromHead(tb, dir, createHead(tb, series))
+}
+
+func createBlockFromHead(tb testing.TB, dir string, head *Head) string {
 	compactor, err := NewLeveledCompactor(context.Background(), nil, log.NewNopLogger(), []int64{1000000}, nil)
 	testutil.Ok(tb, err)
 
